from urlparse import urlunsplit #@@
from mext.uri.parse import urlsplit, unquote, parse_qsl, split_netloc
from mext.uri.encode import dict_to_query, urlquote
from mext.uri.schemes import *


__all__ = ['Uri', 'RelativeUri', 'AbsoluteUri']


# marker for passive attrs from multiple-representation attr sets
# (for ex. _query and _query_dict)
OTHER = object()


#@@ consider making Uris immutable


class Uri(object):
    def __init__(self, *args):
        if len(args) == 1:
            args = urlsplit(unicode(args[0]))
        else:
            args += (None,)*(5-len(args))
        self.scheme, self.netloc, self.path, self.query, self.fragment = args
        cls = scheme_cls_map.get(self.scheme, AbsoluteUri)
        if not isinstance(self, cls):
            if not issubclass(cls, self.__class__):
                raise TypeError("%r is not a subclass of %r" % (cls, self.__class__))
            self.__class__ = cls
        self._post_init()

##     def _post_init(self):
##         pass
##

    @property
    def parts(self):
        return (self.scheme, self.netloc, self.path, self.query, self.fragment)

    def __str__(self):
        return urlunsplit(self.parts) #@@@@

    def __repr__(self):
        return self.__class__.__name__ + '(%s)' % repr(str(self))

    @property
    def path_with_query(self):
        path = self.path
        if self.query:
            path += '?' + self.query
        return path

    path_qs = path_with_query

    @classmethod
    def from_parts(cls, scheme=None, netloc=None, path=None, query=None, fragment=None):
        return cls(scheme, netloc, path, query, fragment)

    def copy(self, **kw):
        r = self.__class__(str(self))
        if kw:
            for k, v in kw.items():
                assert hasattr(r, k)
                setattr(r, k, v)
        return r

    def join(self, ref):
        if isinstance(ref, unicode):
            ref = urlquote(ref, '') #@@?
        if not ref:
            return self
        ref = Uri(ref)
        if ref.is_absolute:
            return ref

        r = Uri(self.scheme, self.netloc, self.path, self.query, ref.fragment)
        if ref.path:
            r.path = joinpath(self.path, ref.path)
            r.query = ref.query
        elif ref.query:
            r.query = ref.query
        return r

    #@@ maybe just use .join('#frag') ?
    def with_fragment(self, fragment):
        return Uri(self.scheme, self.netloc, self.path, self.query, fragment)

    def as_dir(self, save_query=False):
        r = self.copy()
        if not r.path.endswith('/'):
            r.path += '/'
        if not save_query:
            r.query = None
            r.fragment = None
        return r

    def defrag(self):
        if self.fragment:
            return self.with_fragment(None)
        else:
            return self


    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return str(self) == str(Uri(other))

    def __ne__(self, other):
        return not (self == other)


    def _get_query(self):
        if self._query is OTHER:
            self._query = dict_to_query(self._query_dict)
            self._query_dict = OTHER
        return self._query
    def _set_query(self, query):
        self._query = query
        self._query_dict = OTHER
    query = property(_get_query, _set_query)


    def _get_query_dict(self):
        if self._query_dict is OTHER:
            self._query_dict = dict(parse_qsl(self._query)) if self._query else {}
            self._query = OTHER
        return self._query_dict
    def _set_query_dict(self, query_dict):
        self._query = OTHER
        self._query_dict = query_dict
    query_dict = property(_get_query_dict, _set_query_dict)

class RelativeUri(Uri):
    is_absolute = False
    def _post_init(self):
        assert not self.scheme and not self.netloc

    def __str__(self):
        if (self.scheme or self.netloc or self.path or self.query):
            return Uri.__str__(self)
        else:
            return '#%s' % (self.fragment or '')

    def __nonzero__(self):
        return bool(self.path or self.query or self.fragment)




class AbsoluteUri(Uri):
    is_absolute = True
    def relative(self, base):
        if not isinstance(base, Uri):
            base = Uri(base)
        if not base.is_absolute:
            raise ValueError("Cannon generate relative links from a relative base")
        elif self.scheme not in scheme_uses_netloc:
            return self
        elif self.scheme == base.scheme and self.netloc == base.netloc:
            parts = [None, None, self.path, self.query, self.fragment]
            if base.path == self.path and base.query == self.query:
                parts[2] = parts[3] = None
            elif '/' in self.path:
                #@@ better relative paths
                #@@ urlparse.non_hierarchical
                base_split = base.path.rsplit('/', 1)
                self_split = self.path.rsplit('/', 1)
                if base_split[0] == self_split[0]:
                    parts[2] = self_split[1]
            return RelativeUri(*parts)
            #assert base.join(ref) == self, repr((base, ref, base.join(ref), self))
        else:
            return self

    def _post_init(self):
        if self.fragment and self.scheme not in scheme_uses_fragment:
            raise ValueError('URI scheme %r does not support fragment identifiers' % self.scheme)
        if not self.path and self.scheme in scheme_requires_path:
            self.path = u'/'
        #super(AbsoluteUri, self)._post_init()


class HttpUrl(AbsoluteUri):
    _default_port = 80

    @property
    def auth(self):
        return split_netloc(self.netloc)[0]

    @property
    def host(self):
        return split_netloc(self.netloc)[1]

    @property
    def port_nodefault(self):
        r = split_netloc(self.netloc)[2]
        if r:
            return int(r)

    @property
    def port(self):
        return self.port_nodefault or self._default_port


class HttpsUrl(HttpUrl):
    _default_port = 443


scheme_cls_map = {
    None: RelativeUri,
    '': RelativeUri,
    'http': HttpUrl,
    'https': HttpsUrl,
}


def joinpath(base, rel):
    if rel.startswith('/'):
        return rel
    rsegs = base.split('/')
    del rsegs[0], rsegs[-1]
    is_dir = True
    for seg in rel.split('/'):
        if seg == '.':
            is_dir = True
        elif seg == '..':
            if not rsegs:
                continue
            is_dir = True
            del rsegs[-1]
        else:
            is_dir = False
            rsegs.append(seg)
    if is_dir:
        rsegs.append('')
    return '/'+'/'.join(rsegs)



# print joinpath('/a/b/c/', '/./')
# print joinpath('/', '..//')
