from string import ascii_letters, digits
from codecs import charmap_decode

__all__ = ['urlquote', 'dict_to_query', 'dict_to_query_ext']

def _new_quote_table(base, safe_chars, unsafe_chars=''):
    r = base[:]
    _alter_safe(r, safe_chars)
    _alter_unsafe(r, unsafe_chars)
    return r

def _alter_safe(table, safe_chars):
    for c in safe_chars:
        if isinstance(c, str):
            i, c = ord(c), unicode(c)
        else:
            i, c = c, unicode(chr(c))
        table[i] = unicode(c)

def _alter_unsafe(table, unsafe_chars):
    for c in unsafe_chars:
        if isinstance(c, str):
            c = ord(c)
        table[c] = _qt_full[c]


# table to quote ALL chars
_qt_full = map(u'%%%02X'.__mod__, xrange(256))

# table to quote all chars but the alphanumerics and chars allowed
# in every scheme according to RFC 1738
_qt_safe = _new_quote_table(_qt_full, "-_.~")
#_qt_safe = _new_quote_table(_qt_full, "$-_.+!*'(),")
_alter_safe(_qt_safe, ascii_letters)
_alter_safe(_qt_safe, digits)

# table to keep all chars possibly allowed in any specific scheme
# if any chars are reserved in a scheme, you will need to take care when
# encoding its components.
#_qt_safe2 = _new_quote_table(_qt_safe, ';/?:@=&')
_qt_safe2 = _new_quote_table(_qt_safe, "!*'();:@&=+$,/?#[]")

# For example, to encode a path component of an HTTP URL,
# you can use _qt_safe2 but with '?' marked as unsafe
#
# Technically '/;?' must be marked as unsafe, but normally
# the path is already encoded with those chars in their real meaning.
#
# When decoding the URI, you need to do it per-component


_qt_cache = {
    # when encoding something witout knowledge of reserved chars,
    # use urlquote(url, None)
    None: _qt_safe,
    # if you know that there are no reserved chars in your scheme,
    # use urlencode(url, '')
    '': _qt_safe2
}


# the fragment should quote with ''
# the query string should quote '#'
# query args and vals should quote with '!urlarg'
# path should quote '?#'
#??netloc


def urlquote(s, unsafe=None):
    if isinstance(s, unicode):
        s = s.encode('utf8')
    try:
        qtable = _qt_cache[unsafe]
    except KeyError:
        qtable = _qt_safe2[:]
        _alter_unsafe(qtable, unsafe)
        _qt_cache[unsafe] = qtable
    return charmap_decode(s, 'strict', qtable)[0]

def make_quoter(unsafe=None):
    if isinstance(unsafe, list):
        qtable = unsafe
    else:
        try:
            qtable = _qt_cache[unsafe]
        except KeyError:
            qtable = _qt_safe2[:]
            _alter_unsafe(qtable, unsafe)
            _qt_cache[unsafe] = qtable

    def do_quote(qtable, s):
        if isinstance(s, unicode):
            s = s.encode('utf8')
        return str(charmap_decode(s, 'strict', qtable)[0])
    return do_quote.__get__(qtable, list)




# escape name or value of a single http query argument
_qt_qval = _new_quote_table(_qt_safe2, '', '&=+#')
_qt_qval[ord(' ')] = u'+'
_qt_cache['!urlarg'] = _qt_qval

_qval = make_quoter(_qt_qval)
_q_path = make_quoter('?')

#@@ support multiple values per key
def dict_to_query(query, _qval=_qval):
    return '&'.join((_qval(k)+'='+_qval(v)) for (k,v) in query.iteritems())

def dict_to_query_ext(query):
    """
        supports {'k': ['a', 'b'], 'k2': 'c'}
    """
    def _iter_pairs(query):
        for (k,v) in query.iteritems():
            if isinstance(v, (list, tuple)):
                for subv in v:
                    yield k,subv
            elif isinstance(v, (str, unicode)):
                yield k, v
            else:
                raise TypeError('v')
    return '&'.join((_qval(k)+'='+_qval(v)) for k,v in _iter_pairs(query))



## class DebugMap(object):
##     def __getitem__(self, key):
##         r = _qt_qval[key]
##         print 'key:', repr(key)
##         print 'res:', repr(r)
##         #return key
##         return r
##
## #print '!', `_qt_qval[83]`
## _qt_qval_debug = DebugMap()
##
##

##
## print _qval('a')
## print _qval(u'a')
## print `_qval('\xFF')`
## #print dict_to_query({'a': 'foo&foo', 'b': 'bar=x'})
##
## t = '\xd0\xba\xd0\xbe\xd0\xb4\xd0\xb8\xd0\xba \xd1\x8e\xd0\xbd\xd0\xb8\xd0\xba\xd0\xbe\xd0\xb4\xd0\xb8\xd0\xba\n\n--\nVersion: 1.0 RC1 dev\nPlatform: Windows-XP-5.1.2600-SP3'
## print `_qval(t)`
##
## u = urlquote('http+1://foo/a b?ab=x/', '')
## print u
## assert urlquote(u, '') == u
##
